#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import multiprocessing as mp
import numpy as np
import scipy.optimize
from scipy.interpolate import interp1d
from losoto.lib_unwrap import unwrap_2d
from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading TEC module.')

def _run_parser(soltab, parser, step):
    soltabOut = parser.getstr( step, 'soltabOut', 'tec000' )
    refAnt = parser.getstr( step, 'refAnt', '')
    maxResidualFlag = parser.getfloat( step, 'maxResidualFlag', 2.5 )
    maxResidualProp = parser.getfloat( step, 'maxResidualProp', 1. )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['soltabOut', 'refAnt', 'maxResidualFlag', 'maxResidualProp'])
    return run(soltab, soltabOut, refAnt, maxResidualFlag, maxResidualProp, ncpu)


def mod(d):
    """ wrap phases to (-pi,pi)"""
    return np.mod(d + np.pi, 2. * np.pi) - np.pi

_noiseweight = None

def tec_merit_brute(dTEC, freq, phases, weight=True):
    """
    Merit function for brute-force grid search.
    Parameters
    ----------
    dTEC: float, dTEC in TECU
    freq: np array of floats, phases in Hz
    phases: phases to fit in rad
    weight: bool, default True. Whether to weight residuals by noise or not

    Returns
    -------
    rms_phase_residual: float, rms of phase residuals
    """
    if weight:
        w = _noiseweight(freq)
        w /= np.mean(w)
        rms_phase_residual = np.std(w * (mod(-8.44797245e9 * dTEC / freq - phases)))
    else:
        rms_phase_residual = np.std(mod(-8.44797245e9 * dTEC / freq - phases))
    return rms_phase_residual


def tec_merit(dTEC, freq, phases, weight=True):
    """
    Merit function for least-square fit.
    Parameters
    ----------
    dTEC: float, dTEC in TECU
    freq: np array of floats, phases in Hz
    phases: phases to fit in rad
    weight: bool, default True. Whether to weight residuals by noise or not

    Returns
    -------
    rms_phase_residuals: array of floats, rms of phase residuals
    """
    if weight:
        w = _noiseweight(freq)
        w /= np.mean(w)
        rms_phase_residuals = w * mod(-8.44797245e9 * dTEC / freq - phases)
    else:
        rms_phase_residuals = mod(-8.44797245e9 * dTEC / freq - phases)
    return rms_phase_residuals


def fit_tec_to_phases(vals, weights, coord, refAnt, maxResidualFlag, maxResidualProp):
    """
    Fit dTEC to phases and frequencies for a range of times.
    Parameters
    ----------
    vals: numpy array of floats, phases to fit in rad. Shape (n_times, n_freq)
    weights: numpy array of floats, phases to fit in rad. Shape (n_times, n_freq)
    coord: dict of coords of current selection. Contains time, freq, ant, (optional: dir)
    refAnt: string, reference antenna
    maxResidualFlag: float, default = 2.5 Maximum residual that is not flagged. 0=don't flag.
    maxResidualProp: float, default = 1. Maximum residual that is propagated. 0=propagate all.

    Returns
    -------
    [fitdTEC, fitweights]: list of numpy array of floats, dTEC restults in TECU / weights
    """
    # Prepare output arrays
    fitdTEC = np.zeros(len(coord['time']))
    fitweights = np.ones(len(coord['time']))  # all unflagged to start
    # skip refAnt and ants in stationconstraint with refAnt
    if coord['ant'] == refAnt or np.all(vals == 0.):
        pass
    else:
        # find flagged times, either fully flagged or less than 10 freq points...
        flagged_t = np.sum(weights, axis=1)
        flagged_t = flagged_t < 10
        flag_frac = np.sum(flagged_t) / len(flagged_t)
        if flag_frac > 0.1:
            logging.warning(f'''Times with less than 10 unflagged freqs: {coord['time'][flagged_t]}: percentage: {flag_frac:.1%}''')

        ranges, Ns = (-0.5, 0.5), 1000  # default range for brute grid minimization, size of grid
        freq = coord['freq'].copy()
        # Iterate times
        for t, (time, phases, w_t) in enumerate(zip(coord['time'],vals,weights)):
            w_t = w_t.astype(bool)
            if sum(w_t) < 10:
                fitdTEC[t] = 0.
                fitweights[t] = 0
                continue
            # brute force to find global minimum
            dTEC_gridsearch = scipy.optimize.brute(tec_merit_brute, ranges=(ranges,), Ns=Ns, args=(freq[w_t], phases[w_t]))[0]
            result, success = scipy.optimize.leastsq(tec_merit, dTEC_gridsearch, args=(freq[w_t], phases[w_t]))
            best_residual = tec_merit_brute(result, freq[w_t], phases[w_t])
            # logging.info(f'result {result} cost {best_residual}')
            fitdTEC[t] = result

            if maxResidualFlag == 0 or best_residual < maxResidualFlag:
                fitweights[t] = 1
                if maxResidualProp == 0 or best_residual < maxResidualProp:
                    ranges = (fitdTEC[t] - 0.05, fitdTEC[t] + 0.05)
                    Ns = 100
                else:
                    ranges = (-0.5, 0.5)
                    Ns = 1000
            else:
                # high residual, flag and reset initial guess
                if 'dir' in coord.keys():
                    logging.warning('Bad solution for ant: ' + coord['ant'] + '; dir: ' + coord['dir'] + ' (time: ' + str(
                        t) + ', resdiual: ' + str(best_residual) + ').')
                else:
                    logging.warning('Bad solution for ant: ' + coord['ant'] + ' (time: ' + str(t) + ', resdiual: ' + str(
                        best_residual) + ').')
                fitweights[t] = 0
                ranges = (-0.5, 0.5)
                Ns = 1000

                # Debug plot
                # doplot = False
                # if doplot and (coord['ant'] == 'RS509LBA' or coord['ant'] == 'RS210LBA') and t % 50 == 0:
                #     print("Plotting")
                #     if not 'matplotlib' in sys.modules:
                #         import matplotlib as mpl
                #
                #         mpl.rc('figure.subplot', left=0.05, bottom=0.05, right=0.95, top=0.95, wspace=0.22, hspace=0.22)
                #         mpl.use("Agg")
                #     import matplotlib.pyplot as plt
                #
                #     fig = plt.figure()
                #     fig.subplots_adjust(wspace=0)
                #     ax = fig.add_subplot(111)
                #
                #     # plot rm fit
                #     plotd = lambda d, freq: -8.44797245e9 * d / freq
                #     ax.plot(freq, plotd(fitresultd[0], freq[:]), "-", color='purple')
                #     ax.plot(freq, mod(plotd(fitresultd[0], freq[:])), ":", color='purple')
                #
                #     # ax.plot(freq, vals[idx,t], '.b' )
                #     # ax.plot(freq, phaseComb + numjumps * 2*np.pi, 'x', color='purple' )
                #     ax.plot(freq, phaseComb, 'o', color='purple')
                #
                #     residual = mod(plotd(fitd[t], freq[:]) - phaseComb)
                #     ax.plot(freq, residual, '.', color='orange')
                #
                #     ax.set_xlabel('freq')
                #     ax.set_ylabel('phase')
                #     # ax.set_ylim(ymin=-np.pi, ymax=np.pi)
                #
                #     logging.warning('Save pic: ' + str(t) + '_' + coord['ant'] + '.png')
                #     plt.savefig(str(t) + '_' + coord['ant'] + '.png', bbox_inches='tight')
                #     del fig
    if 'dir' in coord.keys():
        logging.info('%s; %s: average tec: %f TECU; std tec: %f TECU' % (coord['ant'], coord['dir'], np.mean(fitdTEC), np.std(fitdTEC))) # prev. factor of 2? why?
    else:
        logging.info('%s: average tec: %f TECU; std tec: %f TECU' % (coord['ant'], np.mean(fitdTEC), np.std(fitdTEC)))
    return [fitdTEC, fitweights]


def run( soltab, soltabOut, refAnt, maxResidualFlag, maxResidualProp, ncpu ):
    """
    Bruteforce TEC extraction from phase solutions.

    Parameters
    ----------
    soltabOut : str, optional
        output table name (same solset), by deault "tec".

    refAnt : str, optional
        Reference antenna, by default the first.

    maxResidualFlag : float, optional
        Max average residual in radians before flagging datapoint, by default 2.5 If 0: no check.

    maxResidualProp : float, optional
        Max average residual in radians before stop propagating solutions, by default 1. If 0: no check.
    """
    # TODO: does not work for global selection, i.e. ant=[...] in parset.
    # define weight function for fitting (global scope)
    global _noiseweight
    # Freq/SEFD estimate for LBA lofar2.0, TODO: move to data dir, add lofar outer/inner
    data_lba_all = np.array([[2.850000000000000000e+07, 2.994030080626945710e+04],
                             [2.903030303030303121e+07, 2.913874315792467314e+04],
                             [2.956060606060606241e+07, 2.848637311863995637e+04],
                             [3.009090909090908989e+07, 2.793457127624304485e+04],
                             [3.062121212121212110e+07, 2.726112344516822850e+04],
                             [3.115151515151515231e+07, 2.656675093925880356e+04],
                             [3.168181818181817979e+07, 2.597196672786273120e+04],
                             [3.221212121212121099e+07, 2.571364929933567691e+04],
                             [3.274242424242424220e+07, 2.546725999266218423e+04],
                             [3.327272727272727340e+07, 2.514084102040380822e+04],
                             [3.380303030303030461e+07, 2.486516674111211978e+04],
                             [3.433333333333333582e+07, 2.467008078555054453e+04],
                             [3.486363636363635957e+07, 2.456891934402090556e+04],
                             [3.539393939393939078e+07, 2.449906041042739525e+04],
                             [3.592424242424242198e+07, 2.441152557008936856e+04],
                             [3.645454545454545319e+07, 2.435284099484204125e+04],
                             [3.698484848484848440e+07, 2.428130362751187931e+04],
                             [3.751515151515151560e+07, 2.412378187290063579e+04],
                             [3.804545454545454681e+07, 2.398172239664983135e+04],
                             [3.857575757575757802e+07, 2.395988451670273935e+04],
                             [3.910606060606060922e+07, 2.430045490772785342e+04],
                             [3.963636363636364043e+07, 2.471031800485042186e+04],
                             [4.016666666666666418e+07, 2.386110364099146318e+04],
                             [4.069696969696969539e+07, 2.353690296084890724e+04],
                             [4.122727272727272660e+07, 2.336803344857854609e+04],
                             [4.175757575757575780e+07, 2.337549122132260163e+04],
                             [4.228787878787878901e+07, 2.282559852902076818e+04],
                             [4.281818181818181276e+07, 2.288857468625183537e+04],
                             [4.334848484848484397e+07, 2.293901290797947513e+04],
                             [4.387878787878787518e+07, 2.279201399700451293e+04],
                             [4.440909090909090638e+07, 2.263575546276593013e+04],
                             [4.493939393939393759e+07, 2.234381817033494008e+04],
                             [4.546969696969696879e+07, 2.201090887195047253e+04],
                             [4.600000000000000000e+07, 2.167511043603119106e+04],
                             [4.653030303030303121e+07, 2.126023280379924108e+04],
                             [4.706060606060606241e+07, 2.094471500899021703e+04],
                             [4.759090909090909362e+07, 2.076310086716620208e+04],
                             [4.812121212121212482e+07, 2.072452147961255469e+04],
                             [4.865151515151514858e+07, 2.072524703336137827e+04],
                             [4.918181818181817979e+07, 2.024623970337267383e+04],
                             [4.971212121212121099e+07, 2.024902352195448111e+04],
                             [5.024242424242424220e+07, 2.042157956576801007e+04],
                             [5.077272727272727340e+07, 2.061233436890421581e+04],
                             [5.130303030303029716e+07, 2.052460108698517070e+04],
                             [5.183333333333332837e+07, 2.044476653198705753e+04],
                             [5.236363636363635957e+07, 2.029609696197773519e+04],
                             [5.289393939393939078e+07, 2.011980530139794064e+04],
                             [5.342424242424242198e+07, 2.000652821604120982e+04],
                             [5.395454545454545319e+07, 1.972491939080056181e+04],
                             [5.448484848484848440e+07, 1.970762288441187411e+04],
                             [5.501515151515151560e+07, 1.963280298568139551e+04],
                             [5.554545454545454681e+07, 1.947178074763434779e+04],
                             [5.607575757575757802e+07, 1.930311130654076624e+04],
                             [5.660606060606060922e+07, 1.930880388910132751e+04],
                             [5.713636363636363298e+07, 1.896554359589192609e+04],
                             [5.766666666666666418e+07, 1.897491931565966297e+04],
                             [5.819696969696969539e+07, 1.899107079252671974e+04],
                             [5.872727272727272660e+07, 1.897387182282819413e+04],
                             [5.925757575757575780e+07, 1.910503545843373649e+04],
                             [5.978787878787878156e+07, 1.940954548637071275e+04],
                             [6.031818181818181276e+07, 1.984135252611626129e+04],
                             [6.084848484848484397e+07, 2.014750239689876980e+04],
                             [6.137878787878787518e+07, 2.045096111245894645e+04],
                             [6.190909090909090638e+07, 2.102328279043795555e+04],
                             [6.243939393939393759e+07, 2.125787547095409172e+04],
                             [6.296969696969696879e+07, 2.165665026710620077e+04],
                             [6.350000000000000000e+07, 2.219586934836382716e+04],
                             [6.403030303030303121e+07, 2.251618565684030182e+04],
                             [6.456060606060606241e+07, 2.256982609767890244e+04],
                             [6.509090909090908617e+07, 2.299487046699088387e+04],
                             [6.562121212121211737e+07, 2.360399101053923732e+04],
                             [6.615151515151514858e+07, 2.452839877523585892e+04],
                             [6.668181818181817979e+07, 2.475335631368393661e+04],
                             [6.721212121212121844e+07, 2.541818455542022275e+04],
                             [6.774242424242424965e+07, 2.632339058027763167e+04],
                             [6.827272727272728086e+07, 2.722260078658507700e+04],
                             [6.880303030303029716e+07, 2.809248882867760403e+04],
                             [6.933333333333332837e+07, 2.869324085220565030e+04],
                             [6.986363636363635957e+07, 2.937062652000980597e+04],
                             [7.039393939393939078e+07, 2.975876253021111552e+04],
                             [7.092424242424242198e+07, 3.023381387615679705e+04],
                             [7.145454545454545319e+07, 3.078095987549442361e+04],
                             [7.198484848484848440e+07, 3.136289409735724985e+04],
                             [7.251515151515151560e+07, 3.194191655745387106e+04],
                             [7.304545454545454681e+07, 3.256773533922632851e+04],
                             [7.357575757575756311e+07, 3.323538538253790466e+04],
                             [7.410606060606059432e+07, 3.408304345233587082e+04],
                             [7.463636363636362553e+07, 3.532438151289376401e+04],
                             [7.516666666666665673e+07, 3.611054891041758674e+04],
                             [7.569696969696968794e+07, 3.677927807149000728e+04],
                             [7.622727272727271914e+07, 3.742928695890052768e+04],
                             [7.675757575757575035e+07, 3.831019856693004840e+04],
                             [7.728787878787878156e+07, 3.942348853846100246e+04],
                             [7.781818181818181276e+07, 3.970818633546015189e+04],
                             [7.834848484848484397e+07, 4.055220533173037984e+04],
                             [7.887878787878787518e+07, 4.174217174788207922e+04],
                             [7.940909090909090638e+07, 4.371878037581651733e+04],
                             [7.993939393939393759e+07, 4.366688385026387550e+04],
                             [8.046969696969696879e+07, 4.441932939676849492e+04],
                             [8.100000000000000000e+07, 4.513826943235301587e+04]])
    sefd_lba_all, wfreq = data_lba_all[:,1], data_lba_all[:,0]
    wgt = 1 / sefd_lba_all
    wgt /= wgt.max()
    _noiseweight = interp1d(wfreq, wgt, bounds_error=False, fill_value=(0, 1))

    logging.info("Find TEC for soltab: "+soltab.name)

    # input check
    solType = soltab.getType()
    if solType != 'phase':
       logging.warning("Soltab type of "+soltab._v_name+" is of type "+solType+", should be phase. Ignoring.")
       return 1

    if 'pol' in soltab.getAxesNames():
        logging.warning("Soltab with pol axis not supported for TEC extraction. Ignoring.")
        return 1
    ants = soltab.getAxisValues('ant')
    if refAnt != '' and refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+ants[0])
        refAnt = ants[0]
    if refAnt == '': refAnt = ants[0]

    # times and ants needs to be complete or selection is much slower
    times = soltab.getAxisValues('time')

    # create new table
    axes = soltab.getAxesNames() # ['time','freq','ant'] or ['time','freq','ant','dir']
    outaxes = axes.copy()
    outaxes.remove('freq') # ['time','ant'] or ['time','ant','dir']
    results_dTEC = np.zeros(shape=tuple(soltab.getAxisLen(axisName) for axisName in outaxes))
    results_w =  np.ones(shape=tuple(soltab.getAxisLen(axisName) for axisName in outaxes))
    solset = soltab.getSolset()
    if 'tec000' in solset.getSoltabNames():
        logging.warning('Soltab tec000 exists. Overwriting...')
        solset.getSoltab('tec000').delete()
    soltabout = solset.makeSoltab(soltype = 'tec', soltabName = soltabOut, axesNames=outaxes, \
                                  axesVals=[soltab.getAxisValues(axisName) for axisName in outaxes], \
                      vals=np.zeros(shape=tuple(soltab.getAxisLen(axisName) for axisName in outaxes)), \
                      weights=np.ones(shape=tuple(soltab.getAxisLen(axisName) for axisName in outaxes)) )
    soltabout.addHistory('Created by TEC operation from %s.' % soltab.name)

    # Collect arguments for pool.map()
    args = []
    selections = []
    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=['time','freq'], weight=True, refAnt=refAnt):
        if len(coord['freq']) < 10:
            logging.error('Delay estimation needs at least 10 frequency channels, preferably distributed over a wide range.')
            return 1
        args.append([vals, weights, coord, refAnt, maxResidualFlag, maxResidualProp])
        selections.append(selection)

    if ncpu == 0:
        ncpu = mp.cpu_count()
    with mp.Pool(ncpu) as pool:
        logging.info('Start TEC fitting.')
        results = pool.starmap(fit_tec_to_phases, args)
    # reorder results
    for selection, result in zip(selections,results):
        selection = tuple([axsel for axsel, ax in zip(selection,axes) if ax in outaxes]) # get rid of selection along freq axis
        results_dTEC[selection] = np.resize(result[0], results_dTEC[selection].shape)
        results_w[selection] = np.resize(result[1], results_dTEC[selection].shape)
    # write results
    soltabout.setValues( results_dTEC )
    soltabout.setValues( results_w, weight=True )

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading POLALIGN module.')

def _run_parser(soltab, parser, step):
    soltabOut = parser.getstr( step, 'soltabOut', 'phasediff' )
    maxResidual = parser.getfloat( step, 'maxResidual', 1. )
    fitOffset = parser.getbool( step, 'fitOffset', False )
    average = parser.getbool( step, 'average', False )
    replace = parser.getbool( step, 'replace', False )
    minFreq = parser.getfloat( step, 'minFreq', 0 )
    refAnt = parser.getstr( step, 'refAnt', '' )

    parser.checkSpelling( step, soltab, ['soltabOut', 'maxResidual', 'fitOffset', 'average', 'replace', 'minFreq', 'refAnt'])
    return run(soltab, soltabOut, maxResidual, fitOffset, average, replace, minFreq, refAnt)


def run( soltab, soltabOut='phasediff', maxResidual=1., fitOffset=False, average=False, replace=False, minFreq=0, refAnt='' ):
    """
    Estimate polarization misalignment as delay.

    Parameters
    ----------
    soltabOut : str, optional
        output table name (same solset), by deault "phasediff".

    maxResidual : float, optional
        Maximum acceptable rms of the residuals in radians before flagging, by default 1. If 0: No check.

    fitOffset : bool, optional
        Assume that together with a delay each station has also a differential phase offset (important for old LBA observations). By default False.    

    average : bool, optional
        Mean-average in time the resulting delays/offset. By default False.    
    
    replace : bool, optional
        replace using smoothed value instead of flag bad data? Smooth must be active. By default, False.

    minFreq : float, optional
        minimum frequency [Hz] to use in estimating the PA. By default, 0 (all freqs).

    refAnt : str, optional
        Reference antenna, by default the first.
    """
    import numpy as np
    import scipy.optimize
    from scipy import stats
    from scipy.ndimage import generic_filter

    logging.info("Finding polarization align for soltab: "+soltab.name)

    delaycomplex = lambda d, freq, y: abs(np.cos(d[0]*freq)  - np.cos(y)) + abs(np.sin(d[0]*freq)  - np.sin(y))
    #delaycomplex = lambda d, freq, y: abs(d[0]*freq  - y)

    solType = soltab.getType()
    if solType != 'phase':
        logging.warning("Soltab type of "+soltab.name+" is of type "+solType+", should be phase. Ignoring.")
        return 1
    
    if refAnt != '' and refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+soltab.getAxisValues('ant')[1])
        refAnt = soltab.getAxisValues('ant')[1]
    if refAnt == '': refAnt = soltab.getAxisValues('ant')[1]

    # times and ants needs to be complete or selection is much slower
    times = soltab.getAxisValues('time')

    # create new table
    solset = soltab.getSolset()
    soltabout = solset.makeSoltab(soltype = soltab.getType(), soltabName = soltabOut, axesNames=soltab.getAxesNames(),
                      axesVals=[soltab.getAxisValues(axisName) for axisName in soltab.getAxesNames()],
                      vals=soltab.getValues(retAxesVals = False), weights=soltab.getValues(weight = True, retAxesVals = False))
    soltabout.addHistory('Created by POLALIGN operation from %s.' % soltab.name)

    if 'XX' in soltab.getAxisValues('pol'): pol = 'XX'
    elif 'RR' in soltab.getAxisValues('pol'): pol = 'RR'
    else:
        logging.error('Cannot reference to known polarisation.')
        return 1

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=['freq','pol','time'], weight=True, refAnt=refAnt):

        # reorder axes
        vals = reorderAxes( vals, soltab.getAxesNames(), ['pol','freq','time'] )
        weights = reorderAxes( weights, soltab.getAxesNames(), ['pol','freq','time'] )

        if 'RR' in coord['pol'] and 'LL' in coord['pol']:
            coord1 = np.where(coord['pol'] == 'RR')[0][0]
            coord2 = np.where(coord['pol'] == 'LL')[0][0]
        elif 'XX' in coord['pol'] and 'YY' in coord['pol']:
            coord1 = np.where(coord['pol'] == 'XX')[0][0]
            coord2 = np.where(coord['pol'] == 'YY')[0][0]

        if (weights == 0.).all() == True:
            logging.warning('Skipping flagged antenna: '+coord['ant'])
            weights[:] = 0
        else:

            fit_delays=[]; fit_offset=[]; fit_weights=[]
            for t, time in enumerate(times):

                # apply flags
                idx       = ( (weights[coord1,:,t] != 0.) & (weights[coord2,:,t] != 0.) & (coord['freq'] > minFreq) )
                freq      = np.copy(coord['freq'])[idx]
                phase1    = vals[coord1,:,t][idx]
                phase2    = vals[coord2,:,t][idx]

                if len(freq) < 30:
                    fit_weights.append(0.)
                    fit_delays.append(0.)
                    fit_offset.append(0.)
                    logging.debug('Not enough unflagged point for the timeslot '+str(t))
                    continue
    
                # if more than 1/2 of chans are flagged
                if (len(idx) - len(freq))/float(len(idx)) > 1/2.:
                    logging.debug('High number of filtered out data points for the timeslot %i: %i/%i' % (t, len(idx) - len(freq), len(idx)) )
    
                phase_diff = phase1 - phase2
                phase_diff = np.mod(phase_diff + np.pi, 2.*np.pi) - np.pi
                phase_diff = np.unwrap(phase_diff)

                A = np.vstack([freq, np.ones(len(freq))]).T
                fitresultdelay = np.linalg.lstsq(A, phase_diff.T)[0]
                # get the closest n*(2pi) to the intercept and refit with only 1 parameter
                if not fitOffset:
                    numjumps = np.around(fitresultdelay[1]/(2*np.pi))
                    A = np.reshape(freq, (-1,1)) # no b
                    phase_diff -= numjumps * 2 * np.pi
                    fitresultdelay = np.linalg.lstsq(A, phase_diff.T)[0]
                    fitresultdelay = [fitresultdelay[0],0.] # set offset to 0 to keep the rest of the script equal

                # fractional residual
                residual = np.mean(np.abs( fitresultdelay[0]*freq + fitresultdelay[1] - phase_diff ))

                fit_delays.append(fitresultdelay[0])
                fit_offset.append(fitresultdelay[1])
                if maxResidual == 0 or residual < maxResidual:
                    fit_weights.append(1.)
                else:       
                    # high residual, flag
                    logging.debug('Bad solution for ant: '+coord['ant']+' (time: '+str(t)+', residual: '+str(residual)+') -> ignoring.')
                    fit_weights.append(0.)

                # Debug plot
                doplot = False
                if doplot and t%100==0 and (coord['ant'] == 'RS310LBA' or coord['ant'] == 'CS301LBA'):
                # if doplot and t%10==0 and (coord['ant'] == 'W04'):
                    if not 'matplotlib' in sys.modules:
                        import matplotlib as mpl
                        mpl.rc('figure.subplot',left=0.05, bottom=0.05, right=0.95, top=0.95,wspace=0.22, hspace=0.22 )
                        mpl.use("Agg")
                    import matplotlib.pyplot as plt

                    fig = plt.figure()
                    fig.subplots_adjust(wspace=0)
                    ax = fig.add_subplot(111)

                    # plot rm fit
                    plotdelay = lambda delay, offset, freq: np.mod( delay*freq + offset + np.pi, 2.*np.pi) - np.pi
                    ax.plot(freq, fitresultdelay[0]*freq + fitresultdelay[1], "-", color='black',  zorder=10, label=r'delay:%f$\nu$ (ns) + %f ' % (fitresultdelay[0]*1e9,fitresultdelay[1]) )

                    ax.plot(freq, np.mod(phase1 + np.pi, 2.*np.pi) - np.pi, 'ob' , label='phase XX/RR')
                    ax.plot(freq, np.mod(phase2 + np.pi, 2.*np.pi) - np.pi, 'og' , label='phase YY/LL' )
                    #ax.plot(freq, np.mod(phase_diff + np.pi, 2.*np.pi) - np.pi, '.', color='purple' )                           
                    ax.plot(freq, phase_diff, '.', color='purple' , label='phase difference')
 
                    residual = np.mod(plotdelay(fitresultdelay[0], fitresultdelay[1], freq)-phase_diff + np.pi,2.*np.pi)-np.pi
                    ax.plot(freq, residual, '.', color='yellow', label='residual')
    
                    ax.set_xlabel('freq')
                    ax.set_ylabel('phase')
                    #ax.set_ylim(ymin=-np.pi, ymax=np.pi)

                    logging.warning('Save pic: '+str(t)+'_'+coord['ant']+'.png')
                    fig.legend()
                    plt.savefig(coord['ant']+'_'+str(t)+'.png', bbox_inches='tight')
                    del fig
            # end cycle in time

            fit_weights = np.array(fit_weights)
            fit_delays = np.array(fit_delays)
            fit_offset = np.array(fit_offset)

            # avg in time
            if average:
                fit_delays_bkp = fit_delays[ fit_weights == 0 ]
                fit_offset_bkp = fit_offset[ fit_weights == 0 ]
                np.putmask(fit_delays, fit_weights == 0, np.nan)
                np.putmask(fit_offset, fit_weights == 0, np.nan)
                fit_delays[:] = np.nanmean(fit_delays)
                # angle mean
                fit_offset[:] = np.angle( np.nansum( np.exp(1j*fit_offset) ) / np.count_nonzero(~np.isnan(fit_offset)) )

                if replace:
                    fit_weights[ fit_weights == 0 ] = 1.
                    fit_weights[ np.isnan(fit_delays) ] = 0. # all the size was flagged cannot estrapolate value
                else:
                    fit_delays[ fit_weights == 0 ] = fit_delays_bkp
                    fit_offset[ fit_weights == 0 ] = fit_offset_bkp

            logging.info('%s: average delay: %f ns (offset: %f)' % ( coord['ant'], np.mean(fit_delays)*1e9, np.mean(fit_offset)))
            for t, time in enumerate(times):
                #vals[:,:,t] = 0.
                #vals[coord1,:,t] = fit_delays[t]*np.array(coord['freq'])/2.
                vals[coord1,:,t] = 0
                phase = np.mod(fit_delays[t]*coord['freq'] + fit_offset[t] + np.pi, 2.*np.pi) - np.pi
                vals[coord2,:,t] = -1.*phase#/2.
                #weights[:,:,t] = 0.
                weights[coord1,:,t] = fit_weights[t]
                weights[coord2,:,t] = fit_weights[t]

        # reorder axes back to the original order, needed for setValues
        vals = reorderAxes( vals, ['pol','freq','time'], [ax for ax in soltab.getAxesNames() if ax in ['pol','freq','time']] )
        weights = reorderAxes( weights, ['pol','freq','time'], [ax for ax in soltab.getAxesNames() if ax in ['pol','freq','time']] )
        soltabout.setSelection(**coord)
        soltabout.setValues( vals )
        soltabout.setValues( weights, weight=True )

    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys, os
import numpy as np
from scipy.interpolate import interp1d
from losoto.h5parm import h5parm, Soltab
from losoto import _version, _logging

_author = "Francesco de Gasperin (astro@voo.it)"

# get input
import argparse
parser = argparse.ArgumentParser(description='Interpolate h5parm files on the fastest varying - '+_author)#, version=_version.__version__
parser.add_argument('h5parmFiles', nargs='+', help='List of h5parms')
parser.add_argument('--insolset', '-s', default='sol000', dest='insolset', help='Input solset name [default: sol000]')
parser.add_argument('--outsolset', '-u', default='sol000', dest='outsolset', help='Output solset name [default: sol000]')
parser.add_argument('--insoltab', '-t', default=None, dest='insoltab', help='Input soltab name (e.g. tabin000) or comma-separated list of soltab names - if not given use all')
parser.add_argument('--interpaxes', '-a', default='time,freq', dest='interpaxes', help='Comma-separated list of axes to interpolate, all other axes need to be equal - default: time,freq')
parser.add_argument('--outh5parm', '-o', default='output.h5', dest='outh5parm', help='Output h5parm name [default: output.h5]')
parser.add_argument('--method', '-m', default='nearest', dest='method', help='see scipy interp1d man, possibilities: linear, nearest, zero, slinear, quadratic, cubic, previous, next. Default: nearest.')
parser.add_argument('--verbose', '-V', '-v', default=False, action='store_true', help='Go Vebose! (default=False)')
parser.add_argument('--clobber', '-c', default=False, action='store_true', help='Replace exising outh5parm file instead of appending to it (default=False)')
args = parser.parse_args()

if len(args.h5parmFiles) < 1:
    parser.print_help()
    sys.exit(1)

# set logs
logger = _logging.Logger('info')
logging = _logging.logger
if args.verbose: logger.set_level("debug")

################################

# check input
if len(args.h5parmFiles) == 1 and (',' in args.h5parmFiles[0] or
                                   ('[' in args.h5parmFiles[0] and
                                    ']' in args.h5parmFiles[0])):
    # Treat input as a string with comma-separated values
    args.h5parmFiles = args.h5parmFiles[0].strip('[]').split(',')
    args.h5parmFiles  = [f.strip() for f in args.h5parmFiles]

# open input
all_soltabs = {} # {'amplitude':[soltab1, soltab2], 'phase':[soltab3]}
all_axes_names = {} # {'amplitude':['dir','time','ant','freq'], 'phase':['dir','time','freq']} # dir is the first by construction
all_axes_vals = {} # {'time':[[1,2,34,5],[1,3,5]],'freq':[[1,2,3,4,5],[1,3,5]]}
pointingNames = []; antennaNames = []; pointingDirections = []; antennaPositions = []
for h5parmFile in args.h5parmFiles:
    logging.info('Working on %s...' % h5parmFile)
    h5 = h5parm(h5parmFile.replace("'",""), readonly=True)
    solset = h5.getSolset(args.insolset)
    insoltabs = solset.getSoltabNames()
    # restrict to requested insoltabs
    if args.insoltab is not None:
        insoltabs = [insoltab for insoltab in insoltabs if insoltab in args.insoltab]
    if len(insoltabs) == 0:
        logging.critical('No soltabs found.')
        sys.exit()
    for insoltab in insoltabs:
        soltab = solset.getSoltab(insoltab)

        # classify soltabs
        typ = soltab.getType()
        if typ in all_soltabs.keys():
            all_soltabs[typ].append( soltab )
        else:
            all_soltabs[typ] = [soltab]

        # classify axes
        axes = soltab.getAxesNames()
        for axis in axes:
            if axis in all_axes_vals.keys():
                all_axes_vals[axis].append( soltab.getAxisValues(axis) )
            else:
                all_axes_vals[axis] = [soltab.getAxisValues(axis)]

        # add axes names for this table typ
        axes = ['dir']+[axis for axis in axes if axis != 'dir']
        if typ in all_axes_names.keys():
            if all_axes_names[typ] != axes:
                logging.error('Table %s has different axes from previously loadded tables of the same type.' % insoltab)
                sys.exit()
        else:
            all_axes_names[typ] = axes


    # collect pointings
    sous = solset.getSou()
    for k,v in list(sous.items()):
        if k not in pointingNames:
            pointingNames.append(k)
            pointingDirections.append(v)

    # collect anntennas
    ants = solset.getAnt()
    for k, v in list(ants.items()):
        if k not in antennaNames:
            antennaNames.append(k)
            antennaPositions.append(v)
   
# Find fastest varying (longest) axis for each axis type
fast_axes_vals = {}
for axis, all_vals in all_axes_vals.items():
    if axis == 'dir':
        fast_axes_vals['dir'] = list(set([val[0] for val in all_vals]))
    else:
        logging.info('Looking for fastest axis in %s...' % axis)
        lens = [len(vals) for vals in all_vals]
        fast_axes_vals[axis] = all_vals[lens.index(max(lens))]
        print(lens,'->',len(fast_axes_vals[axis]))
        if axis not in ['time','freq'] and any([len(fast_axes_vals[axis]) != x for x in lens]):
            logging.error('Only time and freq can be resampled, not %s.' % axis)
            sys.exit()
 
# open output
if os.path.exists(args.outh5parm) and args.clobber:
    os.remove(args.outh5parm)
h5Out = h5parm(args.outh5parm, readonly = False)
# create solset (and add all antennas and directions of other solsets)
solsetOut = h5Out.makeSolset(args.outsolset)
sourceTable = solsetOut.obj._f_get_child('source')
antennaTable = solsetOut.obj._f_get_child('antenna')
antennaTable.append(list(zip(*(antennaNames,antennaPositions))))
sourceTable.append(list(zip(*(pointingNames,pointingDirections))))

for typ, soltabs in all_soltabs.items():
    logging.info('Creating new table of type: %s...' % typ)
    shape = [len(fast_axes_vals[axis]) for axis in all_axes_names[typ]]
    print('Shape: ', shape, ' - ', all_axes_names[typ])
    if typ == 'amplitude':
        new_vals = np.ones(shape=shape)
    else:
        new_vals = np.zeros(shape=shape)
    new_weights = np.ones(shape=shape)

    for soltab in soltabs:
        vals = soltab.getValues(retAxesVals=False)
        weights = soltab.getValues(retAxesVals=False, weight=True)
        #vals[np.isnan(vals)] = 0 # avoid nans (these are flagged data)
        for axis_idx, axis in enumerate(soltab.getAxesNames()):
            if axis not in ['time','freq']: continue
            # get axies
            axis_vals = soltab.getAxisValues(axis)
            new_axis_vals = fast_axes_vals[axis]

            if len(axis_vals) < len(new_axis_vals):
                if len(axis_vals) == 1:
                    pass # numpy should do the proper casting
                else:
                    # interpolate on faster varying axis
                    f = interp1d(axis_vals, vals, kind=args.method, axis=axis_idx, copy=True, fill_value='extrapolate', assume_sorted=True)
                    vals = f(new_axis_vals)
                    fw = interp1d(axis_vals, weights, kind=args.method, axis=axis_idx, copy=True, fill_value='extrapolate', assume_sorted=True)
                    weights = fw(new_axis_vals)

        # remove direciton axis
        idx_axis_dir = soltab.getAxesNames().index('dir') # position of the direction axis
        vals = np.squeeze(vals, axis=idx_axis_dir)
        weights = np.squeeze(weights, axis=idx_axis_dir)
        # find direction
        idx_dir = fast_axes_vals['dir'].index(soltab.getAxisValues('dir'))
        # add to global value
        if typ == 'amplitude':
            new_vals[idx_dir] *= vals
        else:
            new_vals[idx_dir] += vals
        new_weights[idx_dir] *= weights

    # write new soltab (default name)
    soltabOut = solsetOut.makeSoltab(typ, axesNames=all_axes_names[typ],
                axesVals=[ fast_axes_vals[axis_name] for axis_name in all_axes_names[typ] ],
                vals=new_vals, weights=new_weights)

logging.info(str(h5Out))
h5Out.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This operation for LoSoTo implement a jumps remover for TEC solutions
# WEIGHH: flag ready

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading TECJUMP module.')

def _run_parser(soltab, parser, step):
    refAnt = parser.getstr( step, 'refAnt', '' )
    soltabError = parser.getstr( step, 'errorTab', '' )
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['refAnt', 'errorTab'])
    return run(soltab, refAnt, soltabError, ncpu)


def _run_antenna(vals, vals_e, vals_init, weights, selection, tec_jump, antname, outQueue):
    import itertools, random

    extend = 1 # number of point to eaxtend each block

    class Block(object):
        """
        Implement blocks of contiguous series of good datapoints
        """
        def __init__(self, jump_idx_init, jump_idx_end, vals, vals_e, tec_jump, type='poly'):
            self.id = int(np.mean([jump_idx_init,jump_idx_end])) # not used
            self.tec_jump = tec_jump # empirically calculated expected jump size
            self.jump_idx_init = jump_idx_init # indef of first element of the block
            self.jump_idx_end = jump_idx_end # index of last element of the block
            self.idx = range(jump_idx_init, jump_idx_end) # list of indexes
            self.idx_exp = range(jump_idx_init-extend, jump_idx_end+extend) # add 2 at the edges. These are used to calculate matches with other blocks
            self.vals = vals
            self.vals_e = vals_e
            self.len = len(vals) # lenght of the block

            # fill expected values outside the block boundaries
            self.expand(type)

        def get_vals(self, idx):
            """
            Return values and errors.
            Idx is already the local index of the block.
            """
            return self.vals_exp[idx], self.vals_e_exp[idx]

        def expand(self, type='nearest', order=1, size=4):
            """
            predict values+err outside the edges
            TODO: for now it's just nearest
            """
            if self.len == 1: type='nearest' # if 1 point, then nearest-neighbor
            if (self.len-1) < order: order = self.len-1 # 1st order needts 2 ponts, 2nd order 3 and so on...
            if self.len < size: size = self.len # avoid using more points than available

            if type == 'nearest':
                self.vals_exp = np.concatenate( \
                        ( np.full((extend), self.vals[0]) ,
                          self.vals,\
                          np.full((extend), self.vals[-1])
                        ) )
                # errors are between 0.1 and 1
                #self.vals_e_exp = np.concatenate( \
                #        ( np.linspace(2,0.5,1) ,
                #          self.vals_e,\
                #          np.linspace(0.5,2,1)
                #        ) )
                self.vals_e_exp = np.concatenate( \
                        ( [1.]*extend ,
                          self.vals_e,\
                          [1.]*extend
                        ) )

            elif type == 'poly':
                vals_init = self.vals[:size]
                vals_e_init = self.vals_e[:size]
                idx_init = self.idx[:size]
                vals_end = self.vals[-1*size:]
                vals_e_end = self.vals_e[-1*size:]
                idx_end = self.idx[-1*size:]
                p_init = np.poly1d(np.polyfit(idx_init, vals_init, order, w=1./vals_e_init))
                p_end = np.poly1d(np.polyfit(idx_end, vals_end, order, w=1./vals_e_end))
                self.vals_exp = np.concatenate( \
                        ( p_init(self.idx_exp[:extend]) ,
                          self.vals,\
                          p_end(self.idx_exp[-1*extend:])
                        ) )
                # errors are between 100% and 200%
                self.vals_e_exp = np.concatenate( \
                        ( np.linspace(2,1,extend) ,
                          self.vals_e,\
                          np.linspace(1,2,extend)
                        ) )

        def jump(self, jump=0):
            """
            Return values of this block after applying a jump
            """
            self.vals_exp += self.tec_jump*jump

    
#    def distance(block1, block2):
#        """
#        Estimate a distance between two blocks taking advantage of blocks predicted edges
#        """
#        # find indexes of vals_exp that are common to both blocks
#        common_idx_block1 = [i for i, x in enumerate(block1.idx_exp) if x in block2.idx_exp]
#        common_idx_block2 = [i for i, x in enumerate(block2.idx_exp) if x in block1.idx_exp]
#        if len(common_idx_block1) == 0: return 0 # no overlapping
#
#        v1, v1_e = block1.get_vals(idx=common_idx_block1)
#        v2, v2_e = block2.get_vals(idx=common_idx_block2)
#
#        num = np.sum( v1_e * v2_e * (1+abs(v1 - v2))**(1/3.))
#        den = np.sum( v1_e * v2_e )
#
#        return num/den
#
#    def global_potential_old(blocks):
#        """
#        Calculate a "potential" of the time serie summing up the distances between blocks
#        TODO: rewrite this to go through each point, not each block
#        """
#        potential = 0
#        for block1, block2 in zip(blocks[:-1], blocks[1:]):
#            potential += distance(block1, block2)
#
#        return potential

    def global_potential(blocks, idxs):
        """
        Calculate a "potential" of the time serie going thorugh each point
        """
        potentials = 0
        for idx in idxs:
            vals = []
            vals_w = []
            for block in blocks:
                if idx in block.idx_exp:
                    idx_block = block.idx_exp.index(idx)
                    vals.append(block.vals_exp[idx_block])
                    vals_w.append(1./block.vals_e_exp[idx_block])

            if len(vals) == 1: continue # skip inside block
            potential_n = 0
            potential_d = 0
            for idx1, idx2 in itertools.combinations(range(len(vals)),2):
                potential_n += (vals_w[idx1] * vals_w[idx2] * 1/(vals[idx1] - vals[idx2])**2. )
                potential_d += vals_w[idx1] * vals_w[idx2]

            potentials += potential_n/potential_d
                    
        return potentials

    ###############################################################
    
    for i in range(1000):

        # find blocks
        vals_diff = np.diff(vals)
        vals_diff = np.concatenate(([100], list(vals_diff), [100])) # add edges
        jumps_idx = np.where(np.abs(vals_diff) > tec_jump*(2/3.))[0] # find jumps

        # no more jumps
        if len(jumps_idx) == 2: break

        # make block objects
        blocks = []
        for jump_idx_init, jump_idx_end in zip(jumps_idx[:-1],jumps_idx[1:]):
            blocks.append( Block(jump_idx_init, jump_idx_end, \
                    vals[jump_idx_init:jump_idx_end],  vals_e[jump_idx_init:jump_idx_end], tec_jump=tec_jump) )

        # move the small blocks first
        lens = [block.len for block in blocks]
        if len(set(lens)) > 1:
            max_len = sorted(set(lens))[1]
            best_block = blocks[random.choice( np.where( np.array(lens) <= max_len )[0] )] # random block from the two smallest length
        else:
            best_block = random.choice( blocks ) # random block

        #best_block = random.choice(blocks) # random block
        potentials = []
        jump_range = [-5,-4,-3,-2,-1,1,2,3,4,5]
        for jump_size in jump_range:
            best_block.jump(jump_size)
            idx_touse = range(best_block.idx_exp[0], best_block.idx_exp[extend+1]) + \
                        range(best_block.idx_exp[-1*extend-1], best_block.idx_exp[-1]+1) # check only $extend idx around this block
            #print('block:',best_block.idx_exp,idx_touse)
            #print ('Best block (',best_block.id,') idx:',best_block.idx_exp)
            potentials.append( global_potential(blocks, set(idx_touse) ) )
            best_block.jump(-1*jump_size) # return to normality

        #print "potentials:", potentials
            
        # find best jump
        idx = potentials.index( max(potentials) )
        best_jump = jump_range[idx]
        logging.debug('(Ant %s - Cycle %i - #jumps: %i) - Best jump (%i) on block: %i (len %i)' % (antname, i, len(jumps_idx), best_jump, best_block.id, best_block.len) )
            
        # recreate vals with the updated block value
        vals[best_block.idx] = best_block.vals + tec_jump*best_jump

        plot = False
        if plot:
            print ("Preapare plot")
            best_block.vals_exp += tec_jump*best_jump
            import matplotlib as mpl
            mpl.use("Agg")
            import matplotlib.pyplot as plt
            fig = plt.figure(figsize=(16, 8))
            ax = fig.add_subplot(111)
            ax.plot(vals_init, 'k.')
            for block in blocks:
                if block is best_block: continue
                #ax.plot(block.idx_exp, block.vals_exp, 'b,')
                ax.errorbar(block.idx_exp[-1:], block.vals_exp[-1:], block.vals_e_exp[-1:]/30., color='blue', ecolor='blue', marker=',', linestyle='')
                ax.errorbar(block.idx_exp[:1], block.vals_exp[:1], block.vals_e_exp[:1]/30., color='blue', ecolor='blue', marker=',', linestyle='')
            #ax.plot(coord['time'], vals, 'r.')
            ax.errorbar(range(len(vals)), vals, vals_e/30., color='green', ecolor='red', marker='.', linestyle='')
            ax.set_xlim(0,len(vals))
            ax.set_ylim(-0.5,0.5)
            fig.savefig('jump_%s_%03i.png' % (antname, i), bbox_inches='tight')
            fig.clf()

    # check that this is the closest value to the global minimum
    # (this assumes that the majority of the points are correct)
    zeros = []; jumps = []
    for jump in range(-100,101):
        #print('TEST jump %i' % jump)
        #print((vals_init - (vals + tec_jump*jump)) )[:10]
        #print( np.where( np.abs(vals_init - (vals + tec_jump*jump)) < 1e-5 )[0] )
        zeros.append( len( np.where( np.abs(vals_init - (vals + tec_jump*jump)) < 1e-5 )[0] ) )
        jumps.append(jump)
    idx = zeros.index( max(zeros) )
    logging.info('%s: Rescaling all values by %i jumps.' % (antname, jumps[idx]) )
    vals += tec_jump*jumps[idx]
    
    outQueue.put([vals, weights, selection])



def run( soltab, refAnt='', soltabError='', ncpu=0 ):
    """
    Remove jumps from TEC solutions.
    WEIGHT: uses the errors.

    Parameters
    ----------
    soltabError : str, optional
        The table name with solution errors. By default it has the same name of soltab with "error" in place of "tec".

    refAnt : str, optional
        Reference antenna for phases. By default None.

    """

    import scipy.ndimage.filters
    import numpy as np
    from scipy.optimize import minimize
    from scipy.interpolate import griddata
    import scipy.cluster.vq as vq

    def getPhaseWrapBase(freqs):
        """
        freqs: frequency grid of the data
        return the step size from a local minima (2pi phase wrap) to the others [0]: TEC, [1]: clock
        """
        freqs = np.array(freqs)
        nF = freqs.shape[0]
        A = np.zeros((nF, 2), dtype=np.float)
        A[:, 1] = freqs * 2 * np.pi * 1e-9
        A[:, 0] = -8.44797245e9 / freqs
        steps = np.dot(np.dot(np.linalg.inv(np.dot(A.T, A)), A.T), 2 * np.pi * np.ones((nF, ), dtype=np.float))
        return steps

    if soltab.getType() != 'tec':
        logging.error('TECJUMP works only on tec solutions.')
        return 1

    if soltabError == '': soltabError = soltab.name.replace('tec','error')
    solset = soltab.getSolset()
    soltab_e = solset.getSoltab(soltabError)
    try:
        seltab_e = solset.getSoltab(soltabError)
    except:
        logging.error('Cannot fine error solution table %s.' % soltabError)
        return 1
    vals_e_all = soltab_e.getValues( retAxesVals = False, weight = False )

    logging.info("Removing TEC jumps from soltab: "+soltab.name)

    ants = soltab.getAxisValues('ant')
    if refAnt != '' and refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+ants[1])
        refAnt = ants[0]
    if refAnt == '': refAnt = None

    # Get the theoretical tec jump
    tec_jump_theory = abs(getPhaseWrapBase([42.308e6,42.308e6+23828e6])[0])

    # Find the average jump on all antennas by averaging all jumps found to be withing 0.5 and 2 times the tec_jump_theory
    vals = soltab.getValues( retAxesVals=False, refAnt=refAnt )
    timeAxis = soltab.getAxesNames().index('time')
    vals = np.swapaxes(vals, 0, timeAxis)
    vals = vals[1,...] - vals[:-1,...] 
    vals = vals[(vals > tec_jump_theory*1) & (vals < tec_jump_theory*1.5)]
    if len(vals) == 0: 
        logging.info('TEC jump - theoretical: %.5f TECU - NO JUMP FOUND' % (tec_jump_theory))
        return 0
    tec_jump = np.nanmedian(vals)

    logging.info('TEC jump - theoretical: %.5f TECU - estimated: %.5f TECU' % (tec_jump_theory, tec_jump))

    mpm = multiprocManager(ncpu, _run_antenna)

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes='time', weight = True, refAnt=refAnt):

        # skip all flagged
        if (weights == 0).all(): continue
        # skip reference
        if (vals[(weights != 0)] == 0).all(): continue

        logging.info('Working on ant: %s' % (coord['ant']))

        # 1d linear interp to fill flagged data
        vals[np.where(weights == 0)] = np.interp(np.where(weights == 0)[0], np.where(weights != 0)[0], vals[np.where(weights != 0)])

        vals_init = np.copy( vals ) # backup for final check
        vals_e = np.squeeze(vals_e_all[selection])
        vals_e[np.where(weights == 0)] = 1.

        mpm.put([vals, vals_e, vals_init, weights, selection, tec_jump, coord['ant']])

    mpm.wait()
    for (vals, weights, selection) in mpm.get():
        # set back to 0 the values for flagged data
        vals[weights == 0] = 0
        soltab.setValues(vals, selection)
    soltab.addHistory('TECJUMP')

    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading STRUCTURE module.')

def _run_parser(soltab, parser, step):
    doUnwrap = parser.getbool( step, 'doUnwrap', False )
    refAnt = parser.getstr( step, 'refAnt', '')
    plotName = parser.getstr( step, 'plotName', '' )
    ndiv = parser.getint( step, 'ndiv', 1 )

    parser.checkSpelling( step, soltab, ['doUnwrap', 'refAnt', 'plotName', 'ndiv'])
    return run(soltab, doUnwrap, refAnt, plotName, ndiv)


def run( soltab, doUnwrap=False, refAnt='', plotName='', ndiv=1 ):
    """
    Find the structure function from phase solutions of core stations.

    Parameters
    ----------
    doUnwrap : bool, optional

    refAnt : str, optional
        Reference antenna, by default the first.

    plotName : str, optional
        Plot file name, by default no plot.

    ndiv : int, optional
        

    """
    import numpy as np
    from losoto.lib_unwrap import unwrap, unwrap_2d

    logging.info("Find structure function for soltab: "+soltab.name)

    # input check
    solType = soltab.getType()
    if solType != 'phase':
       logging.warning("Soltab type of "+soltab._v_name+" is of type "+solType+", should be phase.")
       return 1

    ants = soltab.getAxisValues('ant')
    if refAnt != '' and refAnt != 'closest' and not refAnt in soltab.getAxisValues('ant', ignoreSelection = True):
        logging.warning('Reference antenna '+refAnt+' not found. Using: '+ants[1])
        refAnt = ants[1]
    if refAnt == '' and doUnwrap:
        logging.error('Unwrap requires reference antenna. Using: '+ants[1])
        refAnt = ants[1]
    if refAnt == '': refAnt = None

    soltab.setSelection(ant='CS*', update=True)

    posAll = soltab.getSolset().getAnt()

    if 'pol' in soltab.getAxesNames(): returnAxes = ['freq','pol','ant','time']
    else: returnAxes = ['freq','ant','time']

    for vals, weights, coord, selection in soltab.getValuesIter(returnAxes=returnAxes, weight=True, refAnt=refAnt):

        # order positions
        pos = np.array([list(posAll[ant]) for ant in coord['ant']])

        # avg pols
        if 'pol' in soltab.getAxesNames():
            # reorder axes
            vals = reorderAxes( vals, soltab.getAxesNames(), ['pol','ant','freq','time'] )
            weights = reorderAxes( weights, soltab.getAxesNames(), ['pol','ant','freq','time'] )

            vals = np.cos(vals) + 1.j * np.sin(vals)
            vals = np.nansum(vals, axis=0)
            vals = np.angle(vals)
            flags = np.array((weights[0]==0)|(weights[1]==0), dtype=bool)
        else:
            # reorder axes
            vals = reorderAxes( vals, soltab.getAxesNames(), ['ant','freq','time'] )
            weights = reorderAxes( weights, soltab.getAxesNames(), ['ant','freq','time'] )
            flags = np.array((weights==0), dtype=bool)

        # unwrap
        if doUnwrap:
            # remove mean to facilitate unwrapping
            for a, ant in enumerate(coord['ant']):
                if not (flags[a,:,:] == True).all() and ant != refAnt:
                    logging.debug('Unwrapping: '+ant)
                    mean = np.angle( np.nanmean( np.exp(1j*vals[a].flatten()) ))
                    vals[a] -= mean
                    vals[a] = np.mod(vals[a]+np.pi, 2*np.pi) - np.pi
                    vals[a,:,:] = unwrap_2d(vals[a,:,:], flags[a,:,:], coord['freq'], coord['time'])
        
        logging.debug('Computing differential values...')
        t1 = np.ma.array( vals, mask=flags ) # mask flagged data
        dph = t1[np.newaxis]-t1[:,np.newaxis] # ant x ant x freq x time
        D = pos[np.newaxis]-pos[:,np.newaxis] # ant x ant x 3
        D2 = np.triu(np.sqrt(np.sum(D**2,axis=-1))) # calc distance and keep only uppoer triangle larger than 0
        myselect = (D2>0)

        if not doUnwrap:
            logging.debug('Re-normalising...')
            dph = np.mod(dph+np.pi, 2*np.pi) - np.pi
            avgdph = np.ma.average(dph, axis=2) # avg in freq (can do because is between -pi and pi)
            #one extra step to remove most(all) phase wraps, phase wraps disturbe the averaging...
            dph = np.remainder(dph - np.ma.average(avgdph,axis=-1)[:,:,np.newaxis,np.newaxis]+np.pi,2*np.pi) + np.ma.average(avgdph,axis=-1)[:,:,np.newaxis,np.newaxis]-np.pi #center around the avg value
        
        logging.debug('Computing sructure function...')
        avgdph = np.ma.average(dph,axis=2) # avg in freq to reduce noise

        variances = []; pars = []
        avgdph = avgdph[...,avgdph.shape[-1]%ndiv:] # remove a few timeslots to make the array divisible by np.split
        avgdph[avgdph.mask] = np.nan
        for i, avgdphSplit in enumerate( np.split(avgdph, ndiv, axis=-1) ):
            variance = np.nanvar(avgdphSplit, axis=-1)*(np.average(coord['freq'])/150.e6)**2 # get time variance and rescale to 150 MHz

            # linear regression
            #A = np.ones((2,D2[myselect].shape[0]),dtype=float)
            #A[1,:] = np.log10(D2[myselect][~variance.mask])
            #par = np.dot(np.linalg.inv(np.dot(A,A.T)),np.dot(A,np.log10(variance[myselect])))
            mask = np.isnan(variance[myselect])
            A = np.vstack([np.log10(D2[myselect][~mask]), np.ones(len(D2[myselect][~mask]))])
            par = np.linalg.lstsq( A.T, np.log10(variance[myselect][~mask]) )[0] 
            S0 = 10**(-1*par[1]/par[0])
            logging.info(r't%i: beta=%.2f - R_diff=%.3f km' % (i, par[0], S0/1.e3))
            variances.append(variance)
            pars.append(par)

        if plotName != '':
            if plotName.split('.')[-1] != 'png': plotName += '.png' # add png

            if not 'matplotlib' in sys.modules:
                import matplotlib as mpl
                mpl.use("Agg")
            import matplotlib.pyplot as plt
    
            fig = plt.figure()
            fig.subplots_adjust(wspace=0)
            ax = fig.add_subplot(111)
            ax1 = ax.twinx()
            
            for i, variance in enumerate(variances):
                if len(variances) > 1:
                    color = plt.cm.jet(i/float(len(variances)-1)) # from 0 to 1
                else: color = 'black'
                ax.plot(D2[myselect]/1.e3,variance[myselect],marker='o',linestyle='', color=color, markeredgecolor='none', label='T')

                # regression
                par = pars[i]
                x = D2[myselect]
                S0 = 10**(-1*par[1]/par[0])
                if color == 'black': color = 'red' # in case of single color, use red line that is more visible
                ax1.plot(x.flatten()/1.e3, par[0]*np.log10(x.flatten()) + par[1], linestyle='-', color=color, label=r'$\beta=%.2f$ - $R_{\rm diff}=%.2f$ km' % (par[0], S0/1.e3))

            ax.set_xlabel('Distance (km)')
            ax.set_ylabel(r'Phase variance @150 MHz (rad$^2$)')
            ax.set_xscale('log')
            ax.set_yscale('log')

            ymin = np.min(variance[myselect])
            ymax = np.max(variance[myselect])
            ax.set_xlim(xmin=0.1,xmax=3)
            ax.set_ylim(ymin,ymax)
            ax1.set_ylim(np.log10(ymin),np.log10(ymax))
            ax1.legend(loc='lower right', frameon=False)
            ax1.set_yticks([])
        
            logging.warning('Save pic: %s' % plotName)
            plt.savefig(plotName, bbox_inches='tight')

    return 0



import os, time, glob
from os.path import dirname, basename, isfile

modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py') and not basename(f).startswith('_')]
del modules
from . import *

class Timer(object):
    """
    context manager used to time the operations
    """

    def __init__(self, log=None, step = 'undef.', operation = 'undef.'):
        """
        log: is a logging istance to print the correct log format
        if nothing is passed, root is used
        """
        if log is None: self.log = logging
        else: self.log = log
        self.step = step
        self.operation = operation

    def __enter__(self):
        self.log.info("--> Starting \'" + self.step + "\' step (operation: " + self.operation + ").")
        self.start = time.time()
        self.startcpu = time.perf_counter()

    def __exit__(self, exit_type, value, tb):

        # if not an error
        if exit_type is None:
            self.log.info("Time for %s step: %i s (cpu: %i s)." % ( self.step, ( time.time() - self.start), (time.perf_counter() - self.startcpu) ))



from __future__ import print_function
from setuptools import setup, Command

import os


# Functions read() and get_version() were copied from Pip package.
# Purpose is to get version info from current package without it
# being installed (which is usually the case when setup.py is run).
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


description = 'LOFAR Solution Tool'
long_description = description
if os.path.exists('README.md'):
    with open('README.md') as f:
        long_description = f.read()


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess
        errno = subprocess.call([sys.executable, 'tools/losoto_test.py'])
        raise SystemExit(errno)


setup(
    name='losoto',
    version=get_version('losoto/_version.py'),
    url='http://github.com/revoltek/losoto/',
    project_urls={
        "Documentation": "https://revoltek.github.io/losoto/",
        "Source": "https://github.com/revoltek/losoto/"
    },
    author='Francesco de Gasperin',
    author_email='astro@voo.it',
    license='GPL',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    platforms='any',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    tests_require=['pytest'],
    install_requires=['numpy>=1.9', 'cython', 'numexpr>=2.0', 'tables>=3.4', 'configparser',
                      'scipy', 'matplotlib', 'python-casacore>=3.0', 'progressbar'],
    scripts=['bin/losoto', 'bin/H5parm_split.py',
             'bin/H5parm2parmdb.py', 'bin/parmdb2H5parm.py', 'bin/killMS2H5parm.py',
             'bin/H5parm_collector.py', 'bin/H5parm_copy.py', 'bin/H5parm_interpolator.py'],
    packages=['losoto', 'losoto.operations', 'losoto.progressbar'],
    test_suite='test',
    cmdclass={'test': PyTest},
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

#from __future__ import division

import itertools, math
import numpy as np
import scipy.fftpack as fft
from losoto._logging import logger as logging

def unwrap_fft(phase, iterations=3):
    """
    Unwrap phase using Fourier techniques.

    For details, see:
    Marvin A. Schofield & Yimei Zhu, Optics Letters, 28, 14 (2003)

    Keyword arguments:
    phase -- array of phase solutions
    iterations -- number of iterations to perform
    """
    puRadius=lambda x : np.roll( np.roll(
          np.add.outer( np.arange(-x.shape[0]/2+1,x.shape[0]/2+1)**2.0,
                        np.arange(-x.shape[1]/2+1,x.shape[1]/2+1)**2.0 ),
          x.shape[1]/2+1,axis=1), x.shape[0]/2+1,axis=0)+1e-9

    idt,dt=np.fft.ifft2,np.fft.fft2
    puOp=lambda x : idt( np.where(puRadius(x)==1e-9,1,puRadius(x)**-1.0)*dt(
          np.cos(x)*idt(puRadius(x)*dt(np.sin(x)))
         -np.sin(x)*idt(puRadius(x)*dt(np.cos(x))) ) )

    def phaseUnwrapper(ip):
       mirrored=np.zeros([x*2 for x in ip.shape])
       mirrored[:ip.shape[0],:ip.shape[1]]=ip
       mirrored[ip.shape[0]:,:ip.shape[1]]=ip[::-1,:]
       mirrored[ip.shape[0]:,ip.shape[1]:]=ip[::-1,::-1]
       mirrored[:ip.shape[0],ip.shape[1]:]=ip[:,::-1]

       return (ip+2*np.pi*
             np.round((puOp(mirrored).real[:ip.shape[0],:ip.shape[1]]-ip)
             /2/np.pi))

    phase2D = phase[:, None]
    i = 0
    if iterations < 1:
        interations = 1
    while i < iterations:
        i += 1
        phase2D = phaseUnwrapper(phase2D)

    return phase2D[:, 0]


def unwrap(phase, window_size=5):
    """
    Unwrap phase by estimating the trend of the phase signal.
    """
    # Allocate result.
    out = np.zeros(phase.shape)

    windowl = np.array([math.fmod(phase[0], 2.0 * math.pi)] * window_size)

    delta = math.fmod(phase[1] - windowl[0], 2.0 * math.pi)
    if delta < -math.pi:
        delta += 2.0 * math.pi
    elif delta > math.pi:
        delta -= 2.0 * math.pi
    windowu = np.array([windowl[0] + delta] * window_size)

    out[0] = windowl[0]
    out[1] = windowu[0]

    meanl = windowl.mean()
    meanu = windowu.mean()
    slope = (meanu - meanl) / float(window_size)

    for i in range(2, len(phase)):
        ref = meanu + (1.0 + (float(window_size) - 1.0) / 2.0) * slope
        delta = math.fmod(phase[i] - ref, 2.0 * math.pi)

        if delta < -math.pi:
            delta += 2.0 * math.pi
        elif delta > math.pi:
            delta -= 2.0 * math.pi

        out[i] = ref + delta

        windowl[:-1] = windowl[1:]
        windowl[-1] = windowu[0]
        windowu[:-1] = windowu[1:]
        windowu[-1] = out[i]

        meanl = windowl.mean()
        meanu = windowu.mean()
        slope = (meanu - meanl) / float(window_size)

    return out


def unwrap_huib( x, window = 10, alpha = 0.01, iterations = 3,
    clip_range = [ 170., 180. ] ):
    """
    Unwrap the x array, if it is shorter than 2*window, use np.unwrap()
    """
    import np as np

    if len(x) < 2*window: return np.unwrap(x)

    xx = np.array( x, dtype = np.float64 )
#   a = zeros( ( window ), dtype = np.float64 )
#   a[ -1 ] = 1.
    if ( len( clip_range ) == 2 ):
        o = clip_range[ 0 ]
        s = ( clip_range[ 1 ] - clip_range[ 0 ] ) / 90.
    a = np.ones( ( window ), dtype = np.float64 ) / float( window )
    xs = xx[ 0 ]
    for j in range( 2 * iterations ):
        for k in range( window, len( x ) ):
            xi = xx[ k - window : k ]
            xp = np.dot( xi, a )
            e = xx[ k ] - xp
            e = np.mod( e + 180., 360. ) - 180.
            if ( len( clip_range ) == 2 ):
                if ( abs( e ) > o ):
                    e = sign( e ) * ( s * degrees( atan( radians( ( abs( e ) - o ) / s ) ) ) + o )
            xx[ k ] = xp + e
#           a = a + xx[ k - window : k ] * alpha * e / 360.
            a = a + xi * alpha * e / ( np.dot( xi, xi ) + 1.e-6 )
        xx = xx[ : : -1 ].copy()
    xx = xx - xx[ 0 ] + xs
    return xx

# This script implements a DCT-based one-step path independent phase
# unwrapping method described in DOI: 10.3390/jimaging1010031 .
# Phase will be unwrapped up to some additive constant.
# - Unwrapping performance depends on the phase map's noise level and C2 smoothness.
# - If zeropadding has to be applied beforehand for FFTs to work, some ugly
#   boundary effects might occur. Best use dims supported by FFTs.
# - Masking out parts of the phase map with zeros also leads to boundary effects.
# - Use it stand alone or as initial guess for preconditioned conjugate gradient method.

# 2D DCT and inverse
def dct2(arr, inverse=False):
    if inverse:
        return np.transpose(fft.idct(np.transpose(fft.idct(arr,norm='ortho')),norm='ortho'))
    else:
        return np.transpose(fft.dct(np.transpose(fft.dct(arr,norm='ortho')),norm='ortho'))

# Laplace operator and inverse
def laplacian(arr, inverse=False):
    # precompute Fourier domain coordinates for the Laplacians of phase_wrapped
    m = np.arange(arr.shape[1])
    n = np.arange(arr.shape[0])
    m,n = np.meshgrid(m,n)
    m = m.astype(float)
    n = n.astype(float)
    m[:,0] = 1/np.sqrt(arr.shape[1])
    n[0,:] = 1/np.sqrt(arr.shape[0])
    coord = m*m+n*n
    m = n = None

    if inverse:
        return dct2(dct2(arr)/coord, inverse=True) # factor -M*N/4/pi**2 omitted
    else:
        return dct2(dct2(arr)*coord, inverse=True) # factor -4*pi**2/M/N omitted

# phase unwrapping, see DOI: 10.3390/jimaging1010031
def unwrap_2d(arr, flags = None, coord_x = None, coord_y = None):
    """
    if flags are specified do interp
    """
    if not flags is None:
        import scipy.interpolate
        if coord_x is None or coord_y is None:
            logging.error('Cannot unwrap with flags and no coordinates.')
            return
        shapeOrig = arr.shape
        grid = np.array([x for x in itertools.product(coord_x,coord_y)])
        arr = np.ndarray.flatten(arr)
        flags = np.ndarray.flatten(flags)
        arr[flags] = scipy.interpolate.griddata(grid[~flags], arr[~flags], grid[flags], 'nearest')
        arr = arr.reshape(shapeOrig)
    return arr + np.round( ( laplacian( np.cos(arr)*laplacian(np.sin(arr)) - np.sin(arr)*laplacian(np.cos(arr)), inverse=True ) - arr ) / 2/np.pi ) * 2*np.pi

#if __name__ == "__main__":
#
#    import matplotlib.pyplot as pl
#    from time import time
#
#    # phase map dims, zeropadding and masking should be avoided if possible
#    N = 1024
#    M = 1024
#    
#    # make some coordinates for the synthetic phase map
#    x = np.arange(N)
#    y = np.arange(M)
#    x,y = np.meshgrid(x,y)
#    x = x / (np.max(x)/(2*np.pi))
#    y = y / (np.max(y)/(2*np.pi))
#    
#    snr = 1 # SNR for the Gaussian noise
#    
#    # some synthetic phase maps
#    #phase_orig = np.sin(x)*x+np.cos(y)*y
#    phase_orig = np.sin(x*y)*x+np.cos(x*x+y*y)*y
#    #phase_orig = (x-np.max(x)/2)**2 + (y-np.max(y)/2)**2
#    x = y = None
#    phase_orig /= np.max(np.abs(phase_orig)) / (np.pi*4.) # scale phase map amplitude
#    phase_orig += np.array( np.random.normal(scale=np.sqrt(0.5/snr),size=(N,M)).astype(np.float32) ) # add noise
#    
#    phase_wrapped = np.arctan2( np.sin(phase_orig), np.cos(phase_orig) ) # wrap phase map to [-pi,+pi]
#    
#    # precompute Fourier domain coordinates for the Laplacians of phase_wrapped
#    m = np.arange(N)
#    n = np.arange(M)
#    m,n = np.meshgrid(m,n)
#    m = m.astype(float)
#    n = n.astype(float)
#    m[:,0] = (1./np.sqrt(M))
#    n[0,:] = (1./np.sqrt(N))
#    coord = m*m+n*n
#    m = n = None
#    
#    #k = 1 # timing iterations
#    t0=time()
#    phase_unwrapped = unwrap_dct(phase_wrapped) # warm-up
#    #for i in range(k):
#    #    phase_unwrapped = unwrap_dct(phase_wrapped)
#    t1 = time()-t0
#    coord = None
#    #print('time:\t%fs'%(t1))
#    
#    #print('norm:', np.abs(phase_unwrapped-phase_orig) )
#    
#    fig, ((ax1,ax2),(ax3,ax4)) = pl.subplots(2,2)
#    im1 = ax1.imshow(phase_orig.__array__())
#    pl.colorbar(im1,ax=ax1)
#    ax1.set_title('original phase, SNR=%i'%snr)
#    im2 = ax2.imshow(phase_wrapped.__array__())
#    pl.colorbar(im2,ax=ax2)
#    ax2.set_title('wrapped phase')
#    im3 = ax3.imshow(phase_unwrapped.__array__())
#    pl.colorbar(im3,ax=ax3)
#    ax3.set_title('unwrapped phase')
#    im4 = ax4.imshow((phase_unwrapped-phase_orig).__array__())
#    pl.colorbar(im4,ax=ax4)
#    ax4.set_title('difference')
#    pl.savefig('test.png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading SPLITLEAK module.')

def _run_parser(soltab, parser, step):
    soltabOutG = parser.getstr( step, 'soltabOutG' ) # no default
    soltabOutD = parser.getstr( step, 'soltabOutD' ) # no default

    parser.checkSpelling( step, soltab, ['soltabOutG','soltabOutD'])
    return run(soltab, soltabOutG, soltabOutD)

def run( soltab, soltabOutG=None, soltabOutD=None):
    """
    Duplicate a table

    Parameters
    ----------
    soltabOutG : str, optional
        Output table name (diagonal component). By default choose next available from table type.

    soltabOutD : str, optional
        Output table name (leakage component). By default choose next available from table type.
    """

    logging.info('Split leakage tables %s -> %s + %s' % (soltab.name, soltabOutG, soltabOutD) )

    if soltab.getType() != 'amplitude' and soltab.getType() != 'phase':
        logging.error('SPLITLEAK can work only on amplitude/phase soltabs. Found: %s.' % soltab.getType())
        return 1
    if not np.all(soltab.getAxisValues('pol') == ['XX', 'XY', 'YX', 'YY']):
        logging.error('Pol in unusual order or not linear: not implemented.')
        return 1

    solset = soltab.getSolset()

    ### G component
    soltabOutG = solset.makeSoltab(soltype = soltab.getType(), soltabName = soltabOutG, axesNames=soltab.getAxesNames(), \
        axesVals=[soltab.getAxisValues(axisName) for axisName in soltab.getAxesNames()], \
        vals=soltab.getValues(retAxesVals = False), weights=soltab.getValues(weight = True, retAxesVals = False))

    # set offdiag to 0
    soltabOutG.setSelection( pol=['XY','YX'] )
    soltabOutG.setValues( 0. )

    ### D component
    soltabOutD = solset.makeSoltab(soltype = soltab.getType(), soltabName = soltabOutD, axesNames=soltab.getAxesNames(), \
        axesVals=[soltab.getAxisValues(axisName) for axisName in soltab.getAxesNames()], \
        vals=soltab.getValues(retAxesVals = False), weights=soltab.getValues(weight = True, retAxesVals = False))

    # divide offdiag by diag, then set diag to 1 (see Hamaker+ 96, appendix D)
    soltabOutD.setSelection(pol=['XX','YY'])
    valsDiag = np.copy(soltabOutD.getValues( retAxesVals = False ))
    if soltab.getType() == 'amplitude':
        soltabOutD.setValues( 1. )
    if soltab.getType() == 'phase':
        soltabOutD.setValues( 0. )

    soltabOutD.setSelection(pol=['XY','YX'])
    valsOffdiag = soltabOutD.getValues( retAxesVals = False )
    if soltab.getType() == 'amplitude':
        soltabOutD.setValues( valsOffdiag/valsDiag )
    if soltab.getType() == 'phase':
        soltabOutD.setValues( valsOffdiag - valsDiag )

    soltabOutG.addHistory('SPLITLEAK: G component of %s' % (soltab.name))
    soltabOutD.addHistory('SPLITLEAK: D component of %s' % (soltab.name))
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This is an example operation for LoSoTo

from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading CLOCKTEC module.')

def _run_parser(soltab, parser, step):
    clocksoltabOut = parser.getstr( step, 'clocksoltabOut', 'clock000' )
    tecsoltabOut = parser.getstr( step, 'tecsoltabOut', 'tec000' )
    offsetsoltabOut = parser.getstr( step, 'offsetsoltabOut', 'phase_offset000' )
    tec3rdsoltabOut = parser.getstr( step, 'tec3rdsoltabOut', 'tec3rd000' )
    flagBadChannels = parser.getbool( step, 'flagBadChannels', True )
    flagCut = parser.getfloat( step, 'flagCut', 5. )
    chi2cut = parser.getfloat( step, 'chi2cut', 3000. )
    combinePol = parser.getbool( step, 'combinePol', False )
    removePhaseWraps = parser.getbool( step, 'removePhaseWraps', True )
    fit3rdorder = parser.getbool( step, 'fit3rdorder', False )
    circular = parser.getbool( step, 'circular', False )
    reverse = parser.getbool( step, 'reverse', False )
    invertOffset = parser.getbool( step, 'invertOffset', False )
    nproc = parser.getint( step, 'nproc', 10 )

    parser.checkSpelling( step, soltab, ['tecsoltabOut', 'clocksoltabOut', 'offsetsoltabOut', 'tec3rdsoltabOut', 'flagBadChannels', 'flagCut', 'chi2cut', 'combinePol', 'removePhaseWraps', 'fit3rdorder', 'circular', 'reverse', 'invertOffset', 'nproc'])
    return run(soltab, tecsoltabOut, clocksoltabOut, offsetsoltabOut, tec3rdsoltabOut, flagBadChannels, flagCut, chi2cut, combinePol, removePhaseWraps, fit3rdorder, circular, reverse, invertOffset, nproc)


def run( soltab, tecsoltabOut='tec000', clocksoltabOut='clock000', offsetsoltabOut='phase_offset000', tec3rdsoltabOut='tec3rd000', flagBadChannels=True, flagCut=5., chi2cut=3000., combinePol=False, removePhaseWraps=True, fit3rdorder=False, circular=False, reverse=False, invertOffset=False, nproc=10 ):
    """
    Separate phase solutions into Clock and TEC.
    The Clock and TEC values are stored in the specified output soltab with type 'clock', 'tec', 'tec3rd'.

    Parameters
    ----------
    flagBadChannels : bool, optional
        Detect and remove bad channel before fitting, by default True.

    flagCut : float, optional


    chi2cut : float, optional


    combinePol : bool, optional
        Find a combined polarization solution, by default False.

    removePhaseWraps : bool, optional
        Detect and remove phase wraps, by default True.

    fit3rdorder : bool, optional
        Fit a 3rd order ionospheric ocmponent (usefult <40 MHz). By default False.

    circular : bool, optional
        Assume circular polarization with FR not removed. By default False.

    reverse : bool, optional
        Reverse the time axis. By default False.

    invertOffset : bool, optional
        Invert (reverse the sign of) the phase offsets. By default False. Set to True
        if you want to use them with the residuals operation.
    """
    import numpy as np
    from ._fitClockTEC import doFit

    logging.info("Clock/TEC separation on soltab: "+soltab.name)

    # some checks
    solType = soltab.getType()
    if solType != 'phase':
       logging.warning("Soltab type of "+soltab.name+" is: "+solType+" should be phase. Ignoring.")
       return 1

    # Collect station properties
    solset = soltab.getSolset()
    station_dict = solset.getAnt()
    stations = soltab.getAxisValues('ant')
    station_positions = np.zeros((len(stations), 3), dtype=np.float)
    for i, station_name in enumerate(stations):
        station_positions[i, 0] = station_dict[station_name][0]
        station_positions[i, 1] = station_dict[station_name][1]
        station_positions[i, 2] = station_dict[station_name][2]

    returnAxes=['ant','freq','pol','time']
    for vals, flags, coord, selection in soltab.getValuesIter(returnAxes=returnAxes,weight=True):

        if len(coord['ant']) < 2:
            logging.error('Clock/TEC separation needs at least 2 antennas selected.')
            return 1
        if len(coord['freq']) < 10:
            logging.error('Clock/TEC separation needs at least 10 frequency channels, preferably distributed over a wide range')
            return 1

        freqs=coord['freq']
        stations=coord['ant']
        times=coord['time']

        # get axes index
        axes=[i for i in soltab.getAxesNames() if i in returnAxes]

        # reverse time axes
        if reverse:
            vals = np.swapaxes(np.swapaxes(vals, 0, axes.index('time'))[::-1], 0, axes.index('time'))
            flags = np.swapaxes(np.swapaxes(flags, 0, axes.index('time'))[::-1], 0, axes.index('time'))

        result=doFit(vals,flags==0,freqs,stations,station_positions,axes,\
                         flagBadChannels=flagBadChannels,flagcut=flagCut,chi2cut=chi2cut,combine_pol=combinePol,removePhaseWraps=removePhaseWraps,fit3rdorder=fit3rdorder,circular=circular,n_proc=nproc)
        if fit3rdorder:
            clock,tec,offset,tec3rd=result
            if reverse:
                clock = clock[::-1,:]
                tec = tec[::-1,:]
                tec3rd = tec3rd[::-1,:]
        else:
            clock,tec,offset=result
            if reverse:
                clock = clock[::-1,:]
                tec = tec[::-1,:]
        if invertOffset:
            offset *= -1.0

        weights=tec>-5
        tec[np.logical_not(weights)]=0
        clock[np.logical_not(weights)]=0
        weights=np.float16(weights)

        if combinePol or not 'pol' in soltab.getAxesNames():
            tf_st = solset.makeSoltab('tec', soltabName = tecsoltabOut,
                             axesNames=['time', 'ant'], axesVals=[times, stations],
                             vals=tec[:,:,0],
                             weights=weights[:,:,0])
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            tf_st = solset.makeSoltab('clock', soltabName = clocksoltabOut,
                             axesNames=['time', 'ant'], axesVals=[times, stations],
                             vals=clock[:,:,0]*1e-9,
                             weights=weights[:,:,0])
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            tf_st = solset.makeSoltab('phase', soltabName = offsetsoltabOut,
                             axesNames=['ant'], axesVals=[stations],
                             vals=offset[:,0],
                             weights=np.ones_like(offset[:,0],dtype=np.float16))
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            if fit3rdorder:
                tf_st = solset.makeSoltab('tec3rd', soltabName = tec3rdsoltabOut,
                                     axesNames=['time', 'ant'], axesVals=[times, stations],
                                     vals=tec3rd[:,:,0],
                                     weights=weights[:,:,0])
        else:
            tf_st = solset.makeSoltab('tec', soltabName = tecsoltabOut,
                             axesNames=['time', 'ant','pol'], axesVals=[times, stations, ['XX','YY']],
                             vals=tec,
                             weights=weights)
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            tf_st = solset.makeSoltab('clock', soltabName = clocksoltabOut,
                             axesNames=['time', 'ant','pol'], axesVals=[times, stations, ['XX','YY']],
                             vals=clock*1e-9,
                             weights=weights)
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            tf_st = solset.makeSoltab('phase', soltabName = offsetsoltabOut,
                             axesNames=['ant','pol'], axesVals=[stations, ['XX','YY']],
                             vals=offset,
                             weights=np.ones_like(offset,dtype=np.float16))
            tf_st.addHistory('CREATE (by CLOCKTECFIT operation)')
            if fit3rdorder:
                tf_st = solset.makeSoltab('tec3rd', soltabName = tec3rdsoltabOut,
                                     axesNames=['time', 'ant','pol'], axesVals=[times, stations, ['XX','YY']],
                                     vals=tec3rd,
                                     weights=weights)
    return 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This operation interpolates the directions of a soltab, adding more directions


import numpy as np
import scipy
from scipy.spatial.distance import cdist
from scipy.interpolate import Rbf
from losoto.lib_operations import *
from losoto._logging import logger as logging

logging.debug('Loading INTERPOLATEDIRECTIONS module.')

# this funct is called by losoto to set parameters and call the real run()
def _run_parser(soltab, parser, step):
    interp_dirs = parser.getarrayfloat2d( step, 'interp_dirs') # no default
    soltabOut = parser.getstr( step, 'soltabOut', '')
    prefix = parser.getstr( step, 'prefix', 'interp_')
    ncpu = parser.getint( '_global', 'ncpu', 0 )

    parser.checkSpelling( step, soltab, ['interp_dirs', 'soltabOut', 'prefix', 'ncpu'])
    return run(soltab, interp_dirs, soltabOut, prefix, ncpu)


def _cartesian_from_radec(x): # TODO is definition of dec correct here?
    return np.array([np.cos(x[:,0])*np.cos(x[:,1]), np.sin(x[:,0])*np.cos(x[:,1]), np.sin(x[:,1])])
    # return np.array([np.cos(x[:,0])*np.sin(x[:,1]), np.sin(x[:,0])*np.sin(x[:,1]), np.cos(x[:,1])])

# def interpolate_directions2d(cal_vals, cal_weights, cal_dirs, dir_ax, interp_dirs, interp_kind, smooth=0.0):
#     """
#     Function to parallelize interpolation. Interpolates and evaluates
#     the interpolation at the new directions for one ant/pol/time etc.
#
#     Parameters
#     ----------
#     cal_vals: array, values of selection
#     weights: weights of values
#     cal_dirs: (2,n) array of floats, ra/dec values of calibrator direc tions
#     dir_ax: int, index of the dir axis.
#     interp_dirs: (2,n) array of floats, ra/dec values of interpolate directions
#     interp_kind: string, 'wrap' or 'lin' dependent on wheter to interpolate phases
#                  (independently in real and imag) or TEC, amp..
#
#     Returns
#     -------
#     new_vals: array of floats, interpolated values of selection at new directions
#     new_weights: array of floats, weights
#     """
#     # approximation: Use euclidean norm in 3d instead of haversine. Every implementation of
#     # haversine I found is much slower than scipy.spatial.distance.cdist for eculidean
#     # to 3d for euclidean approximation. 2d-interpolating in ra-dec has the issue of wraps.
#
#     in_shape = np.shape(cal_vals) # restore shape to reconstruct in output
#     # swap direction axis to first position
#     cal_vals, cal_weights = np.swapaxes(cal_vals, 0, dir_ax), np.swapaxes(cal_weights, 0, dir_ax)
#     # flatten along all other directions
#     cal_vals, cal_weights = cal_vals.reshape(, 0, dir_ax), np.swapaxes(cal_weights, 0, dir_ax)
#     new_weights = np.ones((len(interp_dirs),cal_vals.shape[-1]))
#     cal_weights[np.isnan(cal_vals)] = 0.0 # make sure all NaNs are flagged
#     cal_vals[np.isnan(cal_vals)] = 0.0 # set flagged values to 0
#     new_weights[:,np.sum(cal_weights, axis=0) < len(cal_weights)] # Set new weights to 0 where at least one old direction is flagged.
#
#     # cal_dirs = _cartesian_from_radec(cal_dirs).T
#     # interp_dirs = _cartesian_from_radec(interp_dirs).T
#     if interp_kind == 'wrap': # for phases, interpolate real and imag.
#         _complex = np.exp(1.j * cal_vals)
#         f_interp_re = Rbf(*cal_dirs.T, _complex.real, norm=_haversine, mode='N-D', smooth=smooth) # multiquadratic, linear
#         f_interp_im = Rbf(*cal_dirs.T, _complex.imag, norm=_haversine, mode='N-D', smooth=smooth)
#         interp_re = f_interp_re(*interp_dirs.T)
#         interp_im = f_interp_im(*interp_dirs.T)
#         new_vals = np.angle(interp_re + 1.j * interp_im)
#     elif interp_kind == 'lin':
#         f_interp = Rbf(*cal_dirs.T, cal_vals, norm=_haversine, mode='N-D', smooth=smooth)
#         new_vals = f_interp(*interp_dirs.T)
#     return new_vals, new_weights

def interpolate_directions3d(cal_vals, cal_weights, cal_dirs, dir_axis, interp_dirs, interp_kind, smooth=0.0):
    """
    Function to parallelize interpolation. Interpolates and evaluates
    the interpolation at the new directions for one ant/pol/time etc.
    This function interpolates values on a unit sphere in 3d. This is faster because
    we can use scipy.staptial.distance.cdist euclidean for this.

    Parameters
    ----------
    cal_vals: array, values of selection, 1st axis: dir, other axes flattened
    weights: array, weights of values,
    cal_dirs: (2,n) array of floats, ra/dec values of calibrator directions
    dir_axis: int, position of dir axis in input values/weights
    interp_dirs: (2,n) array of floats, ra/dec values of interpolate directions
    interp_kind: string, 'wrap' or 'lin' dependent on whether to interpolate phases
                 (independently in real and imag) or TEC, amp..
    smooth: float, smooth-parameter for interpolation. For phases, a value between 1e-3 and 0.0
        seems to work well. For TEC, less smoothing should be necessary.

    Returns
    -------
    new_vals: array of floats, interpolated values of selection at new directions
    new_weights: array of floats, weights
    """
    # approximation: Use euclidean norm in 3d instead of haversine. Every implementation of
    # haversine I found is much slower than scipy.spatial.distance.cdist for eculidean
    # to 3d for euclidean approximation. 2d-interpolating in ra-dec has the issue of wraps
    # and no cdist implementation.

    # First reshape the data: move direction axis to first position and flatten all other dims.
    shape_in = np.shape(cal_vals) # restore initial shape at the end
    # get shape with dir at first pos
    shape_swap = list(shape_in)
    shape_swap[0], shape_swap[dir_axis] = shape_swap[dir_axis], shape_swap[0]

    # swap dir axis to firs position
    cal_vals, cal_weights = np.swapaxes(cal_vals, 0, dir_axis), np.swapaxes(cal_weights, dir_axis, 0)
    # flatten along other axes
    cal_vals = np.reshape(cal_vals, (shape_swap[0], np.product(shape_swap[1:])))
    cal_weights = np.reshape(cal_weights, (shape_swap[0], np.product(shape_swap[1:])))

    new_weights = np.ones((len(interp_dirs),cal_vals.shape[-1]))
    cal_weights[np.isnan(cal_vals)] = 0.0 # make sure all NaNs are flagged
    cal_vals[np.isnan(cal_vals)] = 0.0 # set flagged values to 0
    new_weights[:,np.sum(cal_weights, axis=0) < len(cal_weights)] = 0 # Set new weights to 0 where at least one old direction is flagged.
    if interp_kind == 'wrap': # for phases, interpolate real and imag.
        _complex = np.exp(1.j * cal_vals)
        f_interp_re = Rbf(*_cartesian_from_radec(cal_dirs), _complex.real, norm='euclidean', mode='N-D', smooth=smooth) # multiquadratic, linear
        f_interp_im = Rbf(*_cartesian_from_radec(cal_dirs), _complex.imag, norm='euclidean', mode='N-D', smooth=smooth)
        interp_re = f_interp_re(*_cartesian_from_radec(interp_dirs))
        interp_im = f_interp_im(*_cartesian_from_radec(interp_dirs))
        new_vals = np.angle(interp_re + 1.j * interp_im)
    elif interp_kind == 'amp': # use absolute value for amps
        f_interp = Rbf(*_cartesian_from_radec(cal_dirs), cal_vals, norm='euclidean', mode='N-D', smooth=smooth)
        new_vals = np.abs(f_interp(*_cartesian_from_radec(interp_dirs)))
    elif interp_kind == 'lin':
        f_interp = Rbf(*_cartesian_from_radec(cal_dirs), cal_vals, norm='euclidean', mode='N-D', smooth=smooth)
        new_vals = f_interp(*_cartesian_from_radec(interp_dirs))
    # now reconstruct the initial ax order
    new_vals = np.reshape(new_vals, (len(interp_dirs), *shape_swap[1:]))
    new_weights = np.reshape(new_weights, (len(interp_dirs), *shape_swap[1:]))
    # and move back dir axis
    new_vals, new_weights = np.swapaxes(new_vals, 0, dir_axis), np.swapaxes(new_weights, dir_axis, 0)

    return new_vals, new_weights


def run( soltab, interp_dirs, soltabOut=None, prefix='interp_', ncpu=0):
    """
    Add interpolated directions to h5parm
    Parameters
    ----------
    interp_dirs : 2d array of floats
        Shape n x 2, contains ra/dec in degree.
        For example: [[ra1,dec1],[ra2,dec2],...]

    soltabOut : string,  optional
        Default: Guess from soltype. If specifically set to input soltab, the input soltab will be overwritten.

    prefix : string, optional, default = "interp_".
        Name prefix of interpolated directions.
    """
    import multiprocessing as mp
    # need scipy commit f0a478c4a4172c4d2225910216de6b62721db161 for multidimensional interp. otherwise slow...
    if scipy.__version__ < '1.4.0':
        raise ImportError('SciPy version >= 1.4.0 is required to support multidimensional rbf interpolation.')

    logging.info("Working on soltab: "+soltab.name)

    soltype = soltab.getType()
    solset = soltab.getSolset()
    soltabOut = None if soltabOut == '' else soltabOut

    # check input
    if soltype == 'phase':
        interp_kind = 'wrap'
    elif soltype == 'amplitude':
        interp_kind = 'amp'
    elif soltype in ['tec', 'rotationmeasure','tec3rd']:
        interp_kind = 'lin'
    else:
        logging.error('Soltab type {} not supported.'.format(soltype))
        return 1
    if not 'dir' in soltab.getAxesNames():
        logging.error('Data without dir axis cannot be interpolated in directions...')
        return 1
    if not 'ant' in soltab.getAxesNames():
        logging.error('Data without ant axis not supported...')
        return 1
    if len(soltab.dir) < 10:
        logging.error('Not enough directions. Use at least ten for interpolation.')
        return 1

    ax_ord = soltab.getAxesNames() # original order
    dir_ax = ax_ord.index('dir')

    # convert to rad
    if interp_dirs.shape == (2,): # if only one direction, add dir axis
        interp_dirs = interp_dirs[np.newaxis]
    interp_dirs = np.deg2rad(interp_dirs)
    # prepare array for interpolated values - concatenate new directions
    val_shape_inpt = list(soltab.val.shape)
    val_shape_interp = val_shape_inpt.copy()
    val_shape_interp[0] = len(interp_dirs) # match direction axis with no of interp dirs
    interp_vals = np.zeros(val_shape_interp)
    interp_weights = np.ones(val_shape_interp)

    # ra/dec of calibrator dirs. Make sure order is the same as in soltab.
    cal_dirs = np.array([soltab.getSolset().getSou()[k] for k in soltab.dir])

    # Collect arguments for parallelization on antennas
    args, selections = [], []
    returnAxes = ax_ord.copy()
    returnAxes.remove('ant')
    for i, (vals, weights, coord, selection) in enumerate(soltab.getValuesIter(returnAxes=returnAxes, weight=True)):
        dir_ax_sel = returnAxes.index('dir')
        args.append([vals, weights, cal_dirs, dir_ax_sel, interp_dirs, interp_kind])
        selections.append(selection)

    #################### DEBUG PLOT ######################
    if False:
        # What to plot?
        antidx = 32
        sel = [100,201]
        import matplotlib.pyplot as plt
        minra, maxra = np.min(cal_dirs[:,0])-0.005, np.max(cal_dirs[:,0])+0.005
        ra_range = np.linspace(minra, maxra,500)
        mindec, maxdec = np.min(cal_dirs[:,1])-0.005, np.max(cal_dirs[:,1])+0.005
        dec_range = np.linspace(mindec, maxdec,500)
        plotdirs = np.meshgrid(ra_range,dec_range)
        plotdirs = np.array((np.array(plotdirs[0]).flatten(),np.array(plotdirs[1]).flatten())).T
        pvals, pweights = args[antidx][0:2]
        pvals, pweights = np.array(pvals)[:,sel], np.array(pweights)[:,sel]
        plotvals = np.array(interpolate_directions3d(pvals, pweights, cal_dirs, dir_ax, plotdirs, interp_kind, smooth=1.e-3))[0,:,0,0,0]
        plotvals = plotvals.reshape((500,500))
        fig = plt.figure(dpi=200)
        plt.xlabel('RA', labelpad=13, fontsize=7.5)
        plt.ylabel('Dec', labelpad=13, fontsize=7.5)
        cmap = plt.set_cmap('jet')
        if soltype == 'phase':
            label = 'phase [rad]'
            vmin, vmax = -np.pi, np.pi
        elif soltype == 'tec':
            label = 'dTEC [TECU]'
            vmin, vmax = None, None
        elif soltype == 'amplitude':
            label = 'amplitude'
            # vmin, vmax = 0, 1
            vmin, vmax = np.min(plotvals), np.max(plotvals)
            # vmin, vmax = np.min(pvals[:,0,0,0]), np.max(pvals[:,0,0,0])
        else:
            label = 'unknown'
            vmin, vmax = None, None
        im = plt.imshow(plotvals, cmap=cmap, extent = [minra,maxra,maxdec,mindec], vmin=vmin, vmax=vmax)
        cb = fig.colorbar(im)
        plt.scatter(*cal_dirs.T, c=pvals[:,0,0,0], cmap=cmap, edgecolors='k', vmin=vmin, vmax=vmax)
        plt.scatter(*cal_dirs.T, c=pvals[:,0,0,0], cmap=cmap, edgecolors='k', vmin=vmin, vmax=vmax)
        plt.scatter(*interp_dirs.T, facecolor="None", marker='D',edgecolors='k')
        cb.set_label(label, fontsize=7)
        cb.ax.tick_params(labelsize=6.5)
        plt.savefig('debug_screen_interp.png', dpi=200, bbox_inches='tight', )
        import sys
        sys.exit()
    # run the interpolation
    ncpu = mp.cpu_count() if ncpu == 0 else ncpu # default use all cores
    with mp.Pool(ncpu) as pool:
        logging.info('Start interpolation.')
        results = pool.starmap(interpolate_directions3d, args)

    # reorder results
    for selection, result in zip(selections,results):
        vals, weights = result
        selection[dir_ax] = slice(None, None, None) # reset dir selection to apply inout selection to output
        # fill output arrays - the purpose of the reshape is to get the degenerate dim for the ant
        interp_vals[tuple(selection)] = vals.reshape(interp_vals[tuple(selection)].shape)
        interp_weights[tuple(selection)] = weights.reshape(interp_vals[tuple(selection)].shape)

    # concatenate existing values and interpolated direction values
    vals = np.concatenate([soltab.val,interp_vals], axis=dir_ax)
    weights = np.concatenate([soltab.weight,interp_weights], axis=dir_ax)

    # set names for the interpolated directions
    interp_dir_names = np.arange(len(interp_dirs)).astype(str)
    interp_dir_names = [prefix + n.zfill(3) for n in interp_dir_names]
    # prepare axes values, append directions
    axes_vals = [soltab.getAxisValues(axisName) for axisName in ax_ord]
    axes_vals[dir_ax] = np.concatenate([axes_vals[dir_ax],interp_dir_names])

    # prepare output - check if soltabOut exists
    if soltabOut in solset.getSoltabNames():
        logging.warning('Soltab {} exists. Overwriting...'.format(soltabOut))
        solset.getSoltab(soltabOut).delete()

    # make soltabOut
    soltabout = solset.makeSoltab(soltype=soltype, soltabName=soltabOut, axesNames=ax_ord, \
                                  axesVals=axes_vals, vals=vals, weights=weights)

    newSources = dict(zip(interp_dir_names, interp_dirs))
    # append interpolated dirs to solset source table
    sourceTable = solset.obj._f_get_child('source')
    for row in sourceTable.iterrows():
        _name, _dir = row['name'].decode(), row['dir']
        if _name in newSources.keys():
            if not np.all(np.isclose(_dir, newSources[_name])):
                logging.debug('Overwrite direction: {}'.format(_name))
                row['dir'] = newSources[_name]
            else:
                logging.debug('Source already in soltab: {}'.format(_name))
            del newSources[_name]

    if len(newSources) > 0:
        logging.debug('Adding to source table: {}'.format(newSources.keys()))
        sourceTable.append(list(zip(newSources.keys(), newSources.values())))
        # sourceTable.append(list(zip(newSources.keys(), newSources.values())))

    # Add CREATE entry to history
    soltabout.addHistory('Created by INTERPOLATEDIRECTIONS operation from %s.' % soltab.name)
    return 0
